require "numru/dcl"

module NumRu
   module AdvancedDCL
      class Parameter
	 def initialize(parent=nil)
	    if(parent)
	       @parent = parent
	    else
	       @parent = Hash.new
	    end
	    @value = Hash.new
	    @typeinfo = Hash.new
	    @set_funcname = Hash.new
	    @get_funcname = Hash.new
	    @set_call_pattern = Hash.new
	    @get_call_pattern = Hash.new
	    @synced = false
	 end
      
	 def init(key, typeinfo, setfunc, getfunc, set_call, get_call)
	    @value[key] = nil if @parent.has_key?(key)
	    @typeinfo[key] = typeinfo
	    @set_funcname[key] = setfunc
	    @get_funcname[key] = getfunc
	    @set_call_pattern[key] = set_call
	    @get_call_pattern[key] = get_call
	 end
      
	 def keys
	    @typeinfo.keys
	 end

	 def has_key?(key)
	    @typeinfo.has_key?(key)
	 end

	 def value_hash
	    @value
	 end

	 def dup
	    copy = Parameter.new(@parent)
	    
	    value = @value.dup
	    type = @typeinfo
	    set = @set_funcname
	    get = @get_funcname
	    setpat = @set_call_pattern
	    getpat = @get_call_pattern
	    
	    copy.instance_eval{@value = value}
	    copy.instance_eval{@typeinfo = type}
	    copy.instance_eval{@set_funcname = set}
	    copy.instance_eval{@get_funcname = get}
	    copy.instance_eval{@set_call_pattern = setpat}
	    copy.instance_eval{@get_call_pattern = getpat}
	    copy
	 end

	 def copyval(other)
	    @value.each_key{|i|
	       self[i] = other[i]
	    }
	    nil
	 end
	    
	 def valueget(key)
	    value = @value[key]
	    if(value == nil)
	       if(@parent.is_a?(Parameter))
		  value = @parent.valueget(key)
		  return value
	       end
	    end
	    value
	 end
      
	 def [](key)
	    val = @value[key]
	    pval = @parent[key]
	    if(val != nil)  
	       return val
	    elsif(pval != nil)
	       return pval
	    else
	       case @get_call_pattern[key]
	       when 0
		  return eval("DCL.#{@get_funcname[key]}(key)")
	       when 1
		  return eval("DCL.#{@get_funcname[key]}")
	       when 2
		  return eval("#{@get_funcname[key]}")
	       when nil
		  return val
	       else
		  raise "inside error of Parameter class: unsupported call pattern"
	       end
	    end
	 end
	 
	 def []=(key, value)
	    if(@typeinfo[key])
	       if(@get_call_pattern[key] == nil && value == nil)
		  @value[key] = nil
		  return
	       end
	       typeinfo = @typeinfo[key]
	       if(typeinfo.is_a?(Class))
		  if(value.is_a?(typeinfo))
		     @value[key] = value
		  elsif(typeinfo == TrueClass)
		     if(value.type == FalseClass)
			@value[key] = value
		     else
			raise ArgumentError, "value of parametrer "+key+" must be true or false"
		     end
		  else
		     raise ArgumentError, "type of value is incorrect : expects #{typeinfo}"
		  end
	       elsif(typeinfo.is_a?(Array))
		  len = typeinfo[0]
		  typeinfo = typeinfo[1]
		  if(typeinfo.is_a?(Class))
		     klass = typeinfo
		     if(value.is_a?(Array)||value.is_a?(NArray))
			if(value.length == len || len == true)
			   value.each{|i|
			      if(!i.is_a?(klass))
				 raise ArgumentError, "type of elements of value is incorrect : expects #{typeinfo}"
			      end
			   }
			   @value[key] = value
			else
			   raise ArgumentError, "length of array is incorrect : expects #{len}"
			end
		     else
			raise ArgumentError, "value of a parameter " + key.to_s + "expects Array or NArray"
		     end
		  else
		     raise ArgumentError, "typeinfo[0] expects Class or Array"
		  end
	       else
		  raise ArgumentError, "typeinfo expects Class or Array"
	       end
	    else
	       raise ArgumentError, "key,"+key.to_s+" is not available"
	    end
	    @synced = false
	    nil
	 end
	 
	 def sync
	    if(@synced == true)
	       return
	    end

	    @value.each{|key, val|
	       if((val==nil) && (@parent.valueget(key)!=nil))
		  val = @parent[key]
	       end
	       if(val!=nil)
		  case @set_call_pattern[key]
		  when 0
		     eval("DCL.#{@set_funcname[key]}(key, val)")
		  when 1
		     eval("DCL.#{@set_funcname[key]}(val)")
		  when 2
		     eval("DCL.#{@set_funcname[key]}(*val)")
		  when 3
		     eval("self.#{@set_funcname[key]}(val)")
		  when nil
		     nil
		  else
		     raise "non supported call-pattern (#{@set_call_pattern[key]}) is called"
		  end
	       end
	    }
	    @synced = true
	    nil
	 end
	 
#------------------------------------------------
      
	 def set(hash, default=nil, backup=nil)
	    if(default)
	       if(default.is_a?(Parameter))
		  if(backup)
		     backup.copyval(default)
		  else
		     backup = default.dup
		  end
	       else
		  raise ArgumentError, "default type mismatch ( must be Parameter class )"
	       end
	    end
	    if(hash.is_a?(Hash))
	       hash.each{|key, val|
		  if(default)
		     backup[key] = default[key]
		  end
		  self[key] = val
	       }
	    end
	    backup
	 end
	 
	 def clear
	    @value = Hash.new
	 end
	 
	 def help(name)
	    str = "\n available parameter list for " + name.to_s + " \n"
	    str = str + " name   current value   valuetype \n"
	    @typeinfo.each{|key, val|
	       value = self[key]
	       valchar = value.inspect.gsub(/\n/,"\n   ")
	       str = str + key + "  " + valchar + "  " + @typeinfo[key].inspect + " \n"
	    }
	    print str
	 end
      end
   end
end
